# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

'''Utility for rendering an interactive HTML results page from ``ic``
(``wp`` and ``tc`` to follow).
'''

import click
import os
import string
import tempfile
import logging
import urllib.request
import hashlib
import json
from io import StringIO
from .core import common
from . import stats


def _read_data_file(name):
    '''Read a file from the bundled 'lmchallenge/data' directory, and return
    the contents as a string.
    '''
    with open(os.path.join(os.path.dirname(__file__), 'data', name)) as f:
        return f.read()


def _download_cache_cdn(url, sha_384):
    '''Download a file from 'url', which should have the SHA384 matching
    'sha_384' (which should be a hex string).
    '''
    root = os.path.join(tempfile.gettempdir(), 'lmc_page')
    if not os.path.isdir(root):
        os.makedirs(root)

    target = os.path.join(root, sha_384)
    if not os.path.isfile(target):
        logging.info('Downloading %s -> %s', url, target)
        urllib.request.urlretrieve(url, target)
        with open(target, 'rb') as f:
            h = hashlib.sha384()
            h.update(f.read())
            if h.hexdigest() != sha_384:
                logging.error('Checksum mismatch between %s, %s',
                              url, target)
                raise IOError('Checksum mismatch between %s, %s:'
                              ' expected %s actual %s',
                              url, target, h.hexdigest(), sha_384)

    with open(target) as f:
        return f.read()


def _get_files():
    '''Returns a dictionary of {KEY: DATA} for all the supplementary js & css
    data files needed to render the standalone html page.
    '''
    # Note: to get the checksums:
    #   wget https://URL -O - | sha384sum
    return dict(
        PAGE=_read_data_file('page.html'),
        LMC_CSS=_read_data_file('ic.css'),
        LMC_JS=_read_data_file('ic.js'),
        BOOTSTRAP_CSS=_download_cache_cdn(
            'https://'
            'maxcdn.bootstrapcdn.com/bootstrap/3.3.6/css/bootstrap.min.css',
            'd6af264c93804b1f23d40bbe6b95835673e2da59057f0c04'
            '01af210c3763665a4b7a0c618d5304d5f82358f1a6933b3b'
        ),
        BOOTSTRAP_JS=_download_cache_cdn(
            'https://'
            'maxcdn.bootstrapcdn.com/bootstrap/3.3.6/js/bootstrap.min.js',
            'd2649b24310789a95f9ae04140fe80e10ae9aeae4e55f5b7'
            'ecf451de3e442eac6cb35c95a8eb677a99c754ff5a27bc52'
        ),
        JQUERY_JS=_download_cache_cdn(
            'https://code.jquery.com/jquery-2.2.4.min.js',
            'ad8fe3bfc98c86a0da6d74a8f940a082a2ad76605f777a82'
            'dbf2afc930cd43a3dc5095dac4ad6d31ea6841d6b8839bc1'
        ),
        D3_JS=_download_cache_cdn(
            'https://cdnjs.cloudflare.com/ajax/libs/d3/3.5.17/d3.min.js',
            '37c10fd189a5d2337b7b40dc5e567aaedfa2a8a53d0a4e9f'
            'd5943e8f6a6ec5ab6706ae24f44f10eafa81718df82cd6e7'
        ),
    )


def _json_dumps_min(data, float_format=''):
    '''Tiny JSON serializer that supports strings, ints, floats, lists
    and dictionaries.
    Compared to json.dumps, allows a format to specified for floating point
    values.
    '''
    out = StringIO()

    def visit(node):
        if node is None:
            out.write('null')
        elif isinstance(node, str):
            out.write(json.dumps(node))
        elif isinstance(node, int):
            out.write(str(node))
        elif isinstance(node, float):
            out.write(format(node, float_format))
        elif isinstance(node, list):
            out.write('[')
            for i, x in enumerate(node):
                if i != 0:
                    out.write(',')
                visit(x)
            out.write(']')
        elif isinstance(node, dict):
            out.write('{')
            for i, k in enumerate(node):
                if i != 0:
                    out.write(',')
                visit(k)
                out.write(':')
                visit(node[k])
            out.write('}')
    visit(data)
    return out.getvalue()


def _process_ic_log(log, opt):
    '''Prepare the IC log for reading with the JS script ic.js.

    Input format: IC log format - list of line results.

    Output format:
      - note that we use abbreviated keys so that the rendered page is small.
        - "t" = "target"
        - "v" = "verbatim"
        - "r" = "results"
        - "w" = "word"
        - "s" = "score"
        - "e" = "error_score"
        - "m" = "language_model_score"

        [[{"t": TARGET, "v": VERBATIM,
           "r": [{"w": WORD, "s": SCORE,
                  "e": ERROR_SCORE, "m": LANGUAGE_MODEL_SCORE}]}...]...]
    '''
    def get_results(candidates):
        scorer = stats.Ic.scorer(candidates, opt['alpha'], opt['oov_penalty'])
        results = ({'w': x[-3],
                    's': scorer(x[-2:]),
                    'e': x[-2],
                    'm': x[-1]} for x in candidates)
        return sorted(results, key=lambda x: x['s'], reverse=True)

    return [[{'t': x['target'],
              'v': x['verbatim'],
              'r': get_results(x['candidates'])}
             for x in line['inputCorrections']]
            for line in log]


@click.command()
@click.argument('log', nargs=-1, type=click.Path(exists=True, dir_okay=False))
@click.option('-a', '--opt-args', type=common.JsonParam(),
              help='Pass these arguments to the pretty program (for ``ic``),'
              ' for example pre-computed using ``lmc ic-opt``.')
@click.option('-v', '--verbose', default=0, count=True,
              help='How much human-readable detail to print to STDERR.')
@click.option('-f', '--float-fmt', default='.4g',
              help='The format of floats in the JSON file (use compact'
              ' representations to save file size).')
def cli(log, opt_args, verbose, float_fmt):
    '''Create an HTML page rendering of a log file.

    Useful for investigating prediction issues. Currently only 'ic' is
    supported.
    '''
    common.verbosity(verbose)

    if len(log) == 0:
        log = '-'
    elif len(log) == 1:
        log = log[0]
    else:
        raise click.ClickException('Cannot handle multiple log files.')

    if opt_args is None:
        raise click.ClickException(
            'Missing optimization args (run `lmc ic-opt` first to get'
            ' optimization settings).')

    # Create a snippet to substitute in, containing the data to be
    # examined
    ic_data = _process_ic_log(common.read_jsonlines(log), opt_args)
    set_data = string.Template(
        '$$(function () { ic_set_data(${IC_DATA}, ${IC_OPT}); });\n'
    ).substitute(
        IC_DATA=_json_dumps_min(ic_data, float_format=float_fmt),
        IC_OPT=_json_dumps_min(opt_args)
    )

    files = _get_files()
    print(string.Template(files['PAGE']).substitute(
        LMC_SETUP=set_data, **files
    ))


__doc__ += common.shell_docstring(cli, 'lmc page')
if __name__ == '__main__':
    cli()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

'''Main entry point into the evaluation utilities.
'''

import click
import itertools
import json
import sys
import random
from .core import common, runner, errors


# Programs

class Wp:
    def __init__(self, targets=True, predictions=False):
        self.targets = targets
        self.predictions = targets and predictions

    @staticmethod
    def contexts_and_targets(line):
        '''Return a generator of (context, target) pairs by parsing a line.
        '''
        for m in common.TOKENIZER.finditer(line):
            start, end = m.span()
            target = m.group()
            yield line[:start], target

    def __call__(self, model, line):
        '''Return a stats object for a single line of wp evaluation.
        '''
        result = []
        for context, target in Wp.contexts_and_targets(line):
            word_result = {}
            if self.targets:
                word_result['target'] = target
            else:
                word_result['targetChars'] = len(target)
            predictions_with_scores = model.predict(context, None)
            if self.predictions:
                word_result['predictions'] = predictions_with_scores
            target_score = None
            for word, score in predictions_with_scores:
                if word == target:
                    target_score = score
                    break
            if target_score is not None:
                word_result['rank'] = sum(
                    target_score <= score
                    for word, score in predictions_with_scores
                )
                word_result['score'] = target_score

            result.append(word_result)

        return {'wordPredictions': result}


class Tc:
    DEFAULT_SELECT_FROM = 3

    def __init__(self, targets=True, select_from=DEFAULT_SELECT_FROM):
        self.targets = targets
        self.select_from = select_from

    def __call__(self, model, line):
        '''Return a stats object for a single line of tc evaluation.
        '''
        result = []
        typing = ''

        def flush_typing():
            nonlocal typing
            if len(typing) != 0:
                result.append({'target': typing}
                              if self.targets else
                              {'targetChars': len(typing)})
                typing = ''
        i = 0
        while i < len(line):
            context = line[:i]
            tail = line[i:]

            predictions = model.predict(context, None)[:self.select_from]
            matches = [(prediction, rank, score)
                       for rank, (prediction, score) in enumerate(predictions)
                       if len(prediction) != 0 and tail.startswith(prediction)]

            if len(matches) == 0:
                # Typing a character
                typing += line[i]
                i += 1
            else:
                # Accepting a prediction
                flush_typing()
                match, match_rank, match_score = max(matches,
                                                     key=lambda m: len(m[0]))
                entry = {'rank': match_rank,
                         'score': match_score}
                if self.targets:
                    entry['target'] = match
                else:
                    entry['targetChars'] = len(match)
                result.append(entry)
                i += len(match)

        flush_typing()
        return {'textCompletions': result}


class Ic:
    DEFAULT_NUM_CANDIDATES = 100

    def __init__(self, vocab, targets=True,
                 num_candidates=DEFAULT_NUM_CANDIDATES,
                 error_config=errors.DEFAULT_CONFIG):
        self.targets = targets
        self.num_candidates = num_candidates
        self.error_config = error_config
        self.search = errors.Search(words=vocab, n=num_candidates)

    def corruption_candidates(self, target):
        '''Corrupt the candidate, and return a pair ``(corruption, input_scores)``
        with error model scores for each candidate.
        '''

        corrupted = errors.corrupt(self.error_config, target)

        candidates = self.search(corrupted)

        # clip off trailing candidates if needed to ensure corrupted & target
        # can be added - this should keep the maximum size of candidates
        # clamped at 'n' and ensuring that it contains 'target' and 'corrupted'
        candidates = set(
            candidates[:(self.num_candidates - (target not in candidates) -
                         (corrupted not in candidates))]
        ) | set([corrupted, target])

        return corrupted, \
            {c: errors.score(self.error_config, corrupted, c)
             for c in candidates}

    def __call__(self, model, line):
        '''Return a stats object for a single line of ic evaluation.
        '''
        result = []
        for context, target in Wp.contexts_and_targets(line):
            corrupted, input_scores = self.corruption_candidates(target)
            lm_scores = dict(model.predict(context, input_scores.keys()))

            word_result = {}
            word_result['score'] = [input_scores[target],
                                    lm_scores.get(target, None)]
            if self.targets:
                word_result['target'] = target
                word_result['verbatim'] = corrupted
                word_result['candidates'] = [
                    [candidate, input_score, lm_scores.get(candidate, None)]
                    for candidate, input_score in input_scores.items()
                ]
            else:
                word_result['targetChars'] = len(target)
                word_result['verbatimMatch'] = (corrupted == target)
                word_result['candidates'] = [
                    [input_score, lm_scores.get(candidate, None)]
                    for candidate, input_score in input_scores.items()
                ]

            result.append(word_result)

        return {'inputCorrections': result}


# Generic code to run the programs

def run(model, task, data, train=False):
    '''Run an evaluation of ``model`` over some lines of evaluation data.

    ``model`` - an instance of ``runner.BaseModel`` to train & predict

    ``task`` - callable ``task(model, text)`` to return a results object
    for a single line

    ``data`` - a sequence of dictionaries containing at least {"text": "..."}

    ``returns`` - a sequence of dictionaries for each line of text
    '''
    lines_by_user = itertools.groupby(
        data, lambda line: line.get('userId')
    )
    for user_id, user_lines in lines_by_user:
        if train:
            model.clear()
            training_chars = 0

        lines_by_time = itertools.groupby(
            user_lines, lambda line: line.get('timestamp', object())
        )
        for timestamp, same_timestamp_lines in lines_by_time:
            training_buffer = []
            for row in same_timestamp_lines:
                result_row = row.copy()
                text = result_row.pop('text')
                result_row.update(task(model, text))

                if train:
                    training_buffer.append(text)
                    result_row['trainingChars'] = training_chars

                yield result_row

            # When multiple lines of text have the same timestamp
            # (e.g. if they came from the same UsageFragment and we
            # don't know the original order they occurred in), we
            # can't train on one line before predicting another, we
            # have to wait until we've predicted for all lines first.
            if train:
                for line in training_buffer:
                    model.train(line)
                    training_chars += len(text)


class InputFormat(common.ParamChoice):
    '''Input handling - text or json,
    '''
    name = 'input_format'
    choices = ['text', 'json']

    @staticmethod
    def text(line):
        return {'text': line.rstrip('\r\n')}

    @staticmethod
    def json(line):
        return json.loads(line.rstrip('\r\n'))


def run_stream(model, task,
               input=sys.stdin, format=InputFormat.text,
               output=sys.stdout, train=False):
    '''As per ``run``, except wraps model in a ``with``, parses
    input & formats output to file streams.
    '''
    with model as enter_model:
        for line in run(model=enter_model,
                        data=map(format, input),
                        task=task,
                        train=train):
            output.write('%s\n' % json.dumps(line, sort_keys=True))


class PredictorSpec(click.ParamType):
    '''Loads a predictor, either from a Python module or a shell command.
    '''

    class PythonModel:
        def __init__(self, ctor):
            self.ctor = ctor

        def __call__(self, options):
            return self.ctor(options.copy())

    class ShellModel:
        def __init__(self, cmd):
            self.cmd = cmd

        def __call__(self, options):
            return runner.ShellModel(self.cmd, options)

    name = 'predictor_spec'

    def convert(self, value, param, ctx):
        if common.is_qualified_name(value):
            return self.PythonModel(common.lookup_qualified_name(value))
        else:
            return self.ShellModel(value)

    def get_metavar(self, param):
        return 'SPEC'


@click.group()
@click.argument('predictor', type=PredictorSpec())
@click.option('-w/-W', '--targets/--no-targets', default=True,
              help='Omit the target words, leaving only the score and rank'
              ' for each target word. This is useful in case you don\'t want'
              ' to disclose the source data, or to reduce size of the output.'
              ' Not compatible with --predictions (in `wp`).')
@click.option('-t', '--train/--no-train', default=False,
              help='Train the model on lines of text after predictions have'
              ' been given for them (and any others with the same timestamp),'
              ' resetting for each userId. This is mainly useful for dynamic'
              ' modelling experiments with the json corpus format.')
@click.option('-f', '--format', default='text', type=InputFormat(),
              help='Format for test data. text is just lines of plain text;'
              ' json is our json-lines corpus format with rows like'
              ' {"text": "A line of text", "timestamp": 1234567,'
              ' "userId": "user1234"}, ordered by userId then timestamp (where'
              ' present; timestamp is optional).')
@click.option('-o', '--options', type=common.JsonParam(), default='{}',
              help='Additional JSON-formatted options to be parsed and'
              ' passed to a Python module predictor, or converted to command'
              ' line arguments for a shell predictor. In the case of shell'
              ' - the arguments are passed as "--key value", unless key'
              ' already starts with a hyphen (in which case just "key value"),'
              ' and an optional list of arguments with the key "positional".'
              ' For example: \'{"abc": 123, "-n" 10,'
              ' "positional": ["hello", "world"]}\' will be passed as'
              ' `<cmd> hello world -n 10 --a 123`.')
@click.pass_context
def cli(ctx, predictor, targets, train, format, options):
    '''Run a challenge for a predictor over some test text.
    Pipe in text to record an evaluation run of a pipeable predictor, on a
    language modelling task.

    Analyse the output by piping it into `lmchallenge.stats` or
    `lmchallenge.pretty`.

    PREDICTOR - either a shell command to run a pipeable predictor
    e.g. "./my-predictor model.lm",
    or the qualified name of a Python class or function
    e.g. "mymodule.MyClass".
    '''
    ctx.obj = ctx.obj or {}
    ctx.obj['run'] = lambda task: run_stream(predictor(options), task,
                                             format=format,
                                             train=train)
    ctx.obj['targets'] = targets


@cli.command('wp')
@click.option('-p', '--predictions/--no-predictions', default=False,
              help='Include the full list of predictions in the result.')
@click.pass_context
def cli_wp(ctx, predictions):
    '''Word Prediction Challenge.
    '''
    ctx.obj['run'](Wp(targets=ctx.obj['targets'], predictions=predictions))


@cli.command('tc')
@click.option('-n', '--select-from', type=click.INT,
              default=Tc.DEFAULT_SELECT_FROM,
              help='Number of predictions to select text completions from.')
@click.pass_context
def cli_tc(ctx, select_from):
    '''Text Completion Challenge.
    '''
    ctx.obj['run'](Tc(targets=ctx.obj['targets'], select_from=select_from))


@cli.command('ic')
@click.argument('vocab', type=click.File('r'))
@click.option('-r', '--seed', default=42,
              help='Random seed to fix (default: fixed), or 0 to get a'
              ' pseudorandom seed from the clock')
@click.option('-n', '--num-candidates', default=Ic.DEFAULT_NUM_CANDIDATES,
              help='Number of candidates to consider for each'
              ' word \'typed\'.')
@click.option('--error-chars', default=errors.DEFAULT_CONFIG['error_chars'],
              help='Set of characters to sample from when adding'
              ' global character errors')
@click.option('-p', '--p-anykey', default=errors.DEFAULT_CONFIG['p_anykey'],
              help='Probability of substituting an \'anykey\' error')
@click.pass_context
def cli_ic(ctx, vocab, seed, num_candidates, error_chars, p_anykey):
    '''Input Correction Challenge.

    VOCAB - File containing a list of all words to consider as correction
    candidates (this should generally be a long list!)
    '''
    if seed != 0:
        random.seed(seed)
    vocab = set(line.rstrip('\r\n') for line in vocab)
    error_config = dict(error_chars=error_chars, p_anykey=p_anykey)
    ctx.obj['run'](Ic(vocab=vocab, targets=ctx.obj['targets'],
                      num_candidates=num_candidates,
                      error_config=error_config))


__doc__ += common.shell_docstring(cli, 'lmc run')
__doc__ += common.shell_docstring(cli_wp, 'lmc run wp')
__doc__ += common.shell_docstring(cli_tc, 'lmc run tc')
__doc__ += common.shell_docstring(cli_ic, 'lmc run ic')
if __name__ == '__main__':
    cli()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

'''Aggregate results from ``wp``, ``tc`` or ``ic`` challenges, and calculate
summary statistics.
'''

import itertools
import json
import click
import logging
import copy
from collections import defaultdict
from .core import common, monoid
from .core.common import fdiv_null, read_jsonlines


# Optimization helpers

class Domain(click.ParamType):
    '''Parses the domain of a simple linear grid search.
    '''
    class Linear:
        '''A linear sweep through a floating point parameter.
        '''
        slots = ['min', 'max', 'n']

        def __init__(self, min, max, n):
            self.min = min
            self.max = max
            self.n = n

        def values(self):
            '''Returns a list of values to try for this parameter.
            '''
            if self.n == 1:
                return [float(self.min + self.max) / 2]
            step = float(self.max - self.min) / (self.n - 1)
            return [self.min + step * x for x in range(self.n)]

        def __repr__(self):
            return ':'.join(map(str, (self.min, self.max, self.n)))

    name = 'domain'

    def convert(self, value, param, ctx):
        if ':' in value:
            min, max, n = value.split(':')
            return Domain.Linear(min=float(min), max=float(max), n=int(n))
        else:
            fix = float(value)
            return Domain.Linear(min=fix, max=fix, n=1)

    def get_metavar(self, param):
        return '"MIN:MAX:N" | VALUE'


# Stats programs

class Wp:
    '''
    Monoids that may be used to accumulate stats, e.g. (two alternatives)::

      f = common.TokenFilter.alphaonly
      Wp(f).monoid.reduce(log)
      stats.aggregate_by(log, Wp(f).monoid.reduce, stats.Partition.user)
    '''

    # Sentence position metrics, max index
    MAX_INDEX = 5

    @staticmethod
    def _sparse_update(ranks, idx, default, update):
        '''Apply the update function to index idx in the ranks list, extending
        it as necessary with copies of 'default'.'''
        if len(ranks) <= idx:
            ranks.extend([copy.deepcopy(default)
                          for _ in range(1 + idx - len(ranks))])
        ranks[idx] = update(ranks[idx])

    @staticmethod
    def _sparse_merge(aggregate, ranks, default, merge):
        '''Merge the 'ranks' list into the 'aggregate' list.'''
        if len(aggregate) < len(ranks):
            aggregate.extend(map(copy.deepcopy,
                                 [default] * (len(ranks) - len(aggregate))))
        for idx, value in enumerate(ranks):
            aggregate[idx] = merge(aggregate[idx], value)
        return aggregate

    @staticmethod
    def _get_row_ranks(xs):
        ranks = []
        for word_info in xs:
            rank = word_info.get('rank')
            if rank is not None:
                assert rank != 0, 'Rank should be 1-based'
                Wp._sparse_update(ranks, rank - 1, 0, lambda x: x + 1)
        return ranks

    @staticmethod
    def _get_row_ranks_per_idx(xs, max_idx=MAX_INDEX-1):
        '''For each sentence/paragraph, compute ranks for each index in the
        sentence separately, up to max_idx. After max_idx all ranks are
        assigned the same index'''
        ranks = defaultdict(list)
        for idx, word_info in enumerate(xs):
            rank = word_info.get('rank')
            if rank is not None:
                assert rank != 0, 'Rank should be 1-based'
                Wp._sparse_update(ranks[min(max_idx, idx)],
                                  rank - 1, 0, lambda x: x + 1)
        return ranks

    @staticmethod
    def _entry_filter(target_filter):
        '''Convert a target filter into an entry filter (which copes with
        missing targets).'''
        return lambda entry: (('target' not in entry) or
                              target_filter(entry['target']))

    @staticmethod
    def _filtering(filter, include):
        '''Return a Monoid that does custom token filtering on a list of
        inputs'''
        return monoid.Monoid(
            identity=monoid._merge_dicts(include.identity, {'skip': 0}),
            lift=lambda xs: monoid._merge_dicts(
                include.lift([x for x in xs if filter(x)]),
                {'skip': sum(1 for x in xs if not filter(x))}),
            combine=lambda a, b: monoid._merge_dicts(
                include.combine(a, b), {'skip': a['skip'] + b['skip']})
        )

    @staticmethod
    def _merge_idx_dicts(a, b):
        '''Each of a and b is a dict with index of word in the sentence mapped to
        ranks of words within the input'''
        return {k: Wp._sparse_merge(copy.copy(a.get(k, [])), b.get(k, []),
                                    0, lambda x, y: x + y)
                for k in set(a).union(b)}

    per_idx_ranking = monoid.Monoid(
        identity=defaultdict(list),
        lift=_get_row_ranks_per_idx.__func__,
        combine=_merge_idx_dicts.__func__
    )

    ranking = monoid.Monoid(
        identity=[],
        lift=_get_row_ranks.__func__,
        combine=lambda a, b: Wp._sparse_merge(copy.copy(a), b,
                                              0, lambda x, y: x + y)
    )

    word_counting = monoid.map_input(len, monoid.summing)

    per_idx_counting = monoid.Monoid(
        identity=[],
        lift=lambda p: ([int(i < len(p)) for i in range(Wp.MAX_INDEX-1)] +
                        [sum(1 for i, _ in enumerate(p)
                             if i >= Wp.MAX_INDEX-1)]),
        combine=lambda a, b: Wp._sparse_merge(copy.copy(a), b,
                                              0, lambda x, y: x + y))

    char_counting = monoid.map_input(
        lambda xs: sum(len(x['target']) if 'target' in x else x['targetChars']
                       for x in xs),
        monoid.summing
    )

    miss_counting = monoid.map_input(
        lambda xs: sum(1 for x in xs if 'rank' not in x),
        monoid.summing
    )

    stats = monoid.keys(
        ranks=ranking,
        words=word_counting,
        chars=char_counting,
        miss=miss_counting,
        idx_ranks=per_idx_ranking,
        idx_counts=per_idx_counting,
    )

    def __init__(self, filter):
        self.monoid = monoid.map_input(
            lambda row: row['wordPredictions'],
            Wp._filtering(Wp._entry_filter(filter), Wp.stats)
        )

    _rank_metrics = {
        'mrr': lambda r: 1.0 / (1+r),
        'hit1': lambda r: r < 1,
        'hit3': lambda r: r < 3,
        'hit10': lambda r: r < 10,
        'hit20': lambda r: r < 20,
        'hit': lambda r: True
    }
    _idx_rank_metrics = {
        'idx{}_hit3'.format(i): (lambda r: r < 3, i) for i in range(MAX_INDEX)
    }

    def present(self, stats):
        '''A function that transforms the results of Wp statistics.'''
        results = stats.copy()
        results['skip'] = stats['skip'] / stats['words']
        results['miss'] = stats['miss'] / stats['words']

        # ranks
        ranks = results.pop('ranks')
        for name, metric in Wp._rank_metrics.items():
            results[name] = sum(metric(idx) * count for idx, count in
                                enumerate(ranks)) / stats['words']

        # # per-idx ranks
        idx_ranks = results.pop('idx_ranks')
        idx_counts = results.pop('idx_counts')
        for name, (metric, metric_idx) in Wp._idx_rank_metrics.items():
            if metric_idx in idx_ranks:
                results[name] = (sum(metric(idx) * count for idx, count in
                                     enumerate(idx_ranks[metric_idx])) /
                                 idx_counts[metric_idx])
        return results


class Tc:
    '''
    Monoids that may be used to accumulate stats, e.g. (two alternatives)::

        f = common.TokenFilter.alphaonly
        Tc(f).monoid.reduce(log)
        stats.aggregate_by(log, Tc(f).monoid.reduce, stats.Partition.user)
    '''

    @staticmethod
    def _stats(token_filter, row):
        predicted = 0
        calls = 0
        chars = 0
        skip_chars = 0
        for info in row['textCompletions']:
            if 'target' in info:
                allowed = sum(map(token_filter, info['target']))
                skip_chars += len(info['target']) - allowed
            else:
                allowed = info['targetChars']
            chars += allowed
            if info.get('rank') is not None:
                predicted += allowed
                # don't penalize for predictions that don't include allowable
                #  characters
                if allowed != 0:
                    calls += 1
            else:
                calls += allowed
        return {'predicted': predicted,
                'calls': calls,
                'chars': chars,
                'skip_chars': skip_chars}

    @staticmethod
    def completions(filter):
        '''Compute completion statistics from a row of the log given a
        character-level filter.

        If entries include target text, then `filter` is applied.
        Note the target text may be missing if tc was asked to exclude them
        from the evaluation results (e.g. because the source data is private).
        In these cases it is as if the filter allows all.
        '''
        return monoid.Monoid(
            identity={'predicted': 0, 'calls': 0, 'chars': 0, 'skip_chars': 0},
            lift=lambda row: Tc._stats(filter, row),
            combine=lambda x, y: {k: x[k] + y[k]
                                  for k in ['predicted', 'calls',
                                            'chars', 'skip_chars']}
        )

    def __init__(self, filter):
        self.monoid = Tc.completions(filter)

    def present(self, stats):
        '''A function that transforms the results of Tc statistics.'''
        results = stats.copy()
        results['kpc'] = stats['calls'] / stats['chars']
        del results['calls']
        results['pcpc'] = stats['predicted'] / stats['chars']
        del results['predicted']
        results['skip'] = stats['skip_chars'] / stats['chars']
        del results['skip_chars']
        return results


class Ic:
    '''
    Monoids that may be used to accumulate stats, e.g. (two alternatives)::

        f = common.TokenFilter.alphaonly
        Ic(f).monoid.reduce(log)
        stats.aggregate_by(log, Ic(f).monoid.reduce, stats.Partition.user)
    '''

    DEFAULT_ALPHA = Domain.Linear(min=0.0, max=1.0, n=11)
    DEFAULT_OOV_PENALTY = Domain.Linear(min=-8, max=0, n=5)

    @staticmethod
    def optimize(data, target='accuracy', filter=common.TokenFilter.all,
                 alpha=DEFAULT_ALPHA, oov_penalty=DEFAULT_OOV_PENALTY):
        '''Optimize ic predictions (interpolating input and lm probabilities)
        over the given data.
        '''
        def evaluate(args):
            program = Ic(filter, **args)
            score = program.present(program.monoid.reduce(data))[target]
            logging.debug('Run: %s = %f, args = %s', target, score, args)
            return score

        test_args = [dict(alpha=a, oov_penalty=p)
                     for p in oov_penalty.values()
                     for a in alpha.values()]

        logging.debug('Searching: %d configurations', len(test_args))

        scores = [(args, evaluate(args)) for args in test_args]
        best_args, best_score = max(scores, key=lambda x: x[1])

        logging.info('Best: %s = %f, args = %s', target, best_score, best_args)

        return best_args

    @staticmethod
    def scorer(candidates, alpha, oov_penalty):
        '''Return a function that computes the score of a candidate score
        (a candidate is a pair [error_score, lm_score]).
        '''
        language_scores = [x[-1] for x in candidates if x[-1] is not None]
        oov_score = oov_penalty + \
            (min(language_scores) if len(language_scores) else 0)

        def score(x):
            error_score = x[0]
            language_score = x[1] if x[1] is not None else oov_score
            return (1 - alpha) * error_score + alpha * language_score

        return score

    @staticmethod
    def rank(word_info, alpha, oov_penalty):
        '''Return the rank of the correct result with the given mixing
        settings.
        Will always be an integer >= 1.
        '''
        score = Ic.scorer(word_info['candidates'], alpha, oov_penalty)
        target_score = score(word_info['score'])
        return sum(target_score <= score(c[-2:])
                   for c in word_info['candidates'])

    @staticmethod
    def before_pass(word_info):
        '''Return true if the original was already correct (a 'pass').
        '''
        return (word_info['verbatimMatch']
                if 'verbatimMatch' in word_info else
                word_info['target'] == word_info['verbatim'])

    def __init__(self, filter, alpha, oov_penalty):
        self.filter = filter
        self.alpha = alpha
        self.oov_penalty = oov_penalty
        self.monoid = monoid.map_input(
            lambda row: row['inputCorrections'],
            monoid.Monoid(
                identity=Ic._identity, lift=self._lift, combine=Ic._combine
            )
        )

    def _lift(self, results):
        pp = 0
        pf = 0
        fp = 0
        ff = 0
        words = 0
        skip = 0
        for word_info in results:
            word = word_info.get('target')
            words += 1
            if word is not None and not self.filter(word):
                skip += 1
            else:
                before_pass = Ic.before_pass(word_info)
                after_pass = 1 == Ic.rank(word_info,
                                          alpha=self.alpha,
                                          oov_penalty=self.oov_penalty)
                pp += before_pass * after_pass
                pf += before_pass * (not after_pass)
                fp += (not before_pass) * after_pass
                ff += (not before_pass) * (not after_pass)

        return dict(pp=pp, pf=pf, fp=fp, ff=ff, skip=skip, words=words)

    _identity = dict(pp=0, pf=0, fp=0, ff=0, skip=0, words=0)

    @staticmethod
    def _combine(x, y):
        return {k: x[k] + y[k]
                for k in ['pp', 'pf', 'fp', 'ff', 'skip', 'words']}

    def present(self, stats):
        '''A function that transforms the results of Ic statistics.'''
        results = stats.copy()

        results['skip'] = stats['skip'] / stats['words']
        pass_pass = results.pop('pp')
        pass_fail = results.pop('pf')
        fail_pass = results.pop('fp')
        fail_fail = results.pop('ff')

        results['accuracy'] = \
            fdiv_null(fail_pass + pass_pass,
                      fail_pass + pass_pass + fail_fail + pass_fail)
        results['corrected_errors'] = \
            fdiv_null(fail_pass, fail_pass + fail_fail)
        results['miscorrected'] = \
            fdiv_null(pass_fail, pass_fail + pass_pass)
        results['improvement'] = \
            fdiv_null(fail_pass - pass_fail, fail_fail + fail_pass)

        return results


# Generic code to run the different stats programs

class Partition(common.ParamChoice):
    '''Partition functions for stats output.
    '''
    @staticmethod
    def all(s):
        '''Group everything into a single partition.'''
        return True

    @staticmethod
    def user(s):
        '''Group results by user.'''
        return s['userId']

    @staticmethod
    def line(s):
        '''Group results by line.'''
        return object()

    name = 'partition'
    choices = ['all', 'user', 'line']


def aggregate_by(stats, aggregate, partition=Partition.all):
    '''Aggregate a list of stats by a partition function.'''
    return [aggregate(list(s))
            for _, s in itertools.groupby(stats, partition)]


class FragmentProgram:
    monoid = monoid.keys(
        lines=monoid.counting,
        users=monoid.map_input(
            lambda row: row.get('userId'), monoid.uniqing),
        training_chars=monoid.map_input(
            lambda row: row.get('trainingChars'), monoid.maxing)
    )

    def present(self, stats):
        results = stats.copy()
        results['users'] = len(stats['users'])
        return results


class MergeProgram:
    def __init__(self, *programs):
        self.programs = programs
        self.monoid = monoid.merge(*(p.monoid for p in self.programs))

    def present(self, stats):
        results = stats
        for program in self.programs:
            results = program.present(results)
        return results


def stats(data, lines=None, opt_args=None,
          filter=common.TokenFilter.all,
          partition=Partition.all):
    '''Find the correct stats package to use, and compute stats for each
    log file.

    ``data`` - data from a log file to evaluate (e.g. read from
    ``common.read_jsonlines``)

    ``lines`` - optional limit to number of lines to process per file

    ``partition`` - optional partition function to aggregate results over

    ``filter`` - optional filter to select only some tokens to consider

    ``returns`` - a list of [stats_per_aggregate] objects
    '''
    stats_program = common.autodetect_log(data, wp=Wp, tc=Tc, ic=Ic)

    if opt_args is None and hasattr(stats_program, 'optimize'):
        # run the optimizer with default settings
        opt_args = stats_program.optimize(data, filter=filter)

    program = MergeProgram(FragmentProgram(),
                           stats_program(filter, **(opt_args or {})))
    return [program.present(a)
            for a in aggregate_by(data, program.monoid.reduce, partition)]


class Output(common.ParamChoice):
    '''How to print the stats output.
    '''
    @staticmethod
    def json(data):
        '''Dump a results set in json format.'''
        return json.dumps(data[0][1] if len(data) == 1 else dict(data),
                          sort_keys=True)

    @staticmethod
    def csv(data):
        '''Dump a dictionary in csv format (very basic)'''
        if len(data[0][1]) != 1:
            raise ValueError(
                'Can only dump CSV output for unpartitioned data'
                ' (number of partitions: %d)' % len(data[0])
            )
        keys = sorted(data[0][1][0].keys())
        header = ['log'] + keys
        lines = [[log] + [str(results[0][k]) for k in keys]
                 for log, results in data]
        return '\n'.join('\t'.join(row) for row in ([header] + lines))

    name = 'output_format'
    choices = ['json', 'csv']


# Script

@click.command()
@click.argument('log', nargs=-1, type=click.Path(exists=True, dir_okay=False))
@click.option('-v', '--verbose', default=0, count=True,
              help='How much human-readable detail to print to STDERR.')
@click.option('-n', '--lines', type=click.INT,
              help='Limit input to this number of lines')
@click.option('-o', '--output', type=Output(), default='json',
              help='Output format.')
@click.option('-p', '--partition', type=Partition(), default='all',
              help='Group the results for aggregation in this way.')
@click.option('-f', '--filter', type=common.TokenFilter(),
              default='alphaemoji',
              help='Only count tokens which match this filter.')
@click.option('-a', '--opt-args', type=common.JsonParam(), multiple=True,
              help='Pass these arguments to the stats program (for ``ic``),'
              ' for example pre-computed using ``lmc ic-opt``.')
def cli(log, verbose, lines, output, opt_args, **args):
    '''Extract summary stats from any of the challenge log files.
    Specify one or more similar log files, or pipe in results.
    '''
    common.verbosity(verbose)
    log = log or ['-']
    results = [(file, stats(read_jsonlines(file, lines), opt_args=opt, **args))
               for file, opt in common.zip_special(log, opt_args)]
    click.echo(output(results))


__doc__ += common.shell_docstring(cli, 'lmc stats')
if __name__ == '__main__':
    cli()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

"""Utilities for corrupting and correcting text, for use with lmchallenge.ic.

Currently the implemented corruption is very simple - corrupt each character
with a low probability of it being replaced by another ASCII letter (or in-word
punctuation character).
"""

import random
import math
import string
import heapq

DEFAULT_CONFIG = dict(
    p_anykey=0.1,
    error_chars=string.ascii_letters + '_-#@'
)


def corrupt(config, word, rand=random):
    """Generate a corrupted version of a word, as if typed by a sloppy typist.
    """
    p_anykey = config['p_anykey']
    error_chars = config['error_chars']
    return ''.join(
        rand.choice(error_chars) if rand.random() < p_anykey else ch
        for ch in word
    )


def score(config, input_word, word):
    """Return an approximate score for this word, given the error model of
    'config'.
    """
    p_anykey = config['p_anykey']
    n_correct = sum(a == b for a, b in zip(input_word, word))
    return n_correct * math.log(1 - p_anykey) + \
        (len(word) - n_correct) * math.log(p_anykey)


class Search:
    """Functor for finding a list of nearby candidates to a corrupted word.
    """
    def __init__(self, words, n):
        self.words = {}
        for w in words:
            words_l = self.words.get(len(w))
            if words_l is None:
                words_l = []
                self.words[len(w)] = words_l
            words_l.append(w)
        self.n = n

    @staticmethod
    def _count_matches(a, b):
        n = 0
        for ch_a, ch_b in zip(a, b):
            if ch_a == ch_b:
                n += 1
        return n

    def __call__(self, input_word):
        return heapq.nlargest(
            self.n, self.words.get(len(input_word), []),
            key=lambda w: Search._count_matches(input_word, w)
        )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

import sys
import subprocess
import shlex

__doc__ = '''
Core LM Challenge APIs 'runner' for LMC. The core 'Model' duck-API is as
follows::

    # Get text completions following a context.
    # input: context - text before prediction (string)
    #        candidates - candidates to consider (list of strings, or None)
    # output: candidate_scores - to follow context (list of (string, float))
    candidate_scores = model.predict(context, candidates)

    # Adapt the model to the user, providing text that has been entered
    # input: text - text from the current user (string)
    model.train(text)

    # Reset all trained state in the model (i.e. we're about to start with
    # a new user)
    model.clear()

    # In addition, the model should be usable in Python's 'with' statement:
    model.__enter__()
    model.__exit__(type, value, traceback)
    '''


class BaseModel:
    '''Base class for implementing the Model API for LM Challenge.
    Subclasses must at least implement
    ``candidates = predict(self, context, candidates)``
    to be a valid predictor.
    '''

    def predict(self, context, candidates):
        raise NotImplementedError

    def train(self, text):
        '''Default implementation: do nothing.'''
        pass

    def clear(self):
        '''Default implementation: do nothing.'''
        pass

    def __enter__(self):
        '''Default implementation.'''
        return self

    def __exit__(self, type, value, traceback):
        '''Default implementation: do nothing.'''
        pass

    def run_loop(self, input_stream=sys.stdin,
                 output_stream=sys.stdout, error_stream=sys.stderr):
        '''Run the model as a pipeable predictor process between
        ``input_stream`` and ``output_stream``. This method does not
        return until ``input_stream`` is closed.
        '''
        for line in input_stream:
            parts = line.strip('\n').split('\t')
            cmd = parts[0]
            if cmd == 'predict':
                context = parts[1] if 2 <= len(parts) else ''
                candidates = parts[2:] if 3 <= len(parts) else None
                results = self.predict(context, candidates)
                response = '\t'.join('%s\t%f' % (candidate, score)
                                     for candidate, score in results) + '\n'
                output_stream.write(response)
            elif cmd == 'train':
                self.train(parts[1])
            elif cmd == 'clear':
                self.clear()
            else:
                error_stream.write('Unrecognized command "' + cmd + '"\n')
            output_stream.flush()


class ShellModel(BaseModel):
    '''Defines a language model that communicates over a Unix pipe'''

    def __init__(self, cmd, opts):
        '''Open a pipe to the model'''
        self.verbose = opts.get('verbose', False)
        cmd_positional = ' '.join(opts.get('positional', []))
        cmd_options = ' '.join(
            '%s %s' % (key if key.startswith('-') else ('--' + key),
                       shlex.quote(str(value)))
            for key, value in opts.items()
            if key != 'positional'
        )
        self.cmd = '%s %s %s' % (cmd, cmd_positional, cmd_options)
        self._debug('$ %s' % self.cmd)
        self.proc = subprocess.Popen(self.cmd, shell=True,
                                     stdin=subprocess.PIPE,
                                     stdout=subprocess.PIPE)
        self.proc_in = open(self.proc.stdin.fileno(), 'w', encoding='utf-8')
        self.proc_out = open(self.proc.stdout.fileno(), 'r', encoding='utf-8')
        self._check_return_code()

    def __exit__(self, type, value, traceback):
        self._close()

    def _debug(self, line):
        if self.verbose:
            sys.stderr.write(line + '\n')

    def _close(self):
        '''Finish with the process & shut it down.'''
        self.proc.communicate()
        self._check_return_code()

    def _check_return_code(self):
        '''Check that our model process hasn't errored out.'''
        rc = self.proc.returncode
        if rc is not None and rc != 0:
            raise subprocess.CalledProcessError(rc, self.cmd)

    def _send_command(self, command):
        '''Issue a tab-delimited command to the model process.'''
        self._debug('#> %s' % (' <TAB> '.join(command)))
        self._check_return_code()
        self.proc_in.write(u'\t'.join(command) + u'\n')
        self.proc_in.flush()

    def predict(self, context, candidates):
        command = ['predict', context]
        if candidates is not None:
            command += candidates
        self._send_command(command)
        self._check_return_code()
        candidates_and_scores = [
            s for s in self.proc_out.readline().strip('\n').split('\t')
        ]
        self._debug('#< %s' % (' <TAB> '.join(candidates_and_scores)))
        pairs = [(candidates_and_scores[i*2],
                  float(candidates_and_scores[i*2+1]))
                 for i in range(len(candidates_and_scores)//2)]
        return pairs

    def train(self, text):
        self._send_command(['train', text])

    def clear(self):
        self._send_command(['clear'])



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

'''LM Challenge - language modelling evaluation suite.
'''

import click
from . import run, stats, pretty, ic_opt, page


@click.group()
def cli():
    '''The main entry point to LMChallenge.
    Use subcommands to perform specific tasks.
    '''
    pass


cli.add_command(run.cli, 'run')
cli.add_command(stats.cli, 'stats')
cli.add_command(pretty.cli, 'pretty')
cli.add_command(ic_opt.cli, 'ic-opt')
cli.add_command(page.cli, 'page')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

import os
import re
import sys
import regex
import click
import emoji
import logging
import json
import itertools
import warnings
import importlib
import itertools as it

TOKENIZER = regex.compile(
    emoji.get_emoji_regexp().pattern +
    """|[\p{L}\p{N}\p{Pc}\p{Pd}'@#]+|[\p{P}\p{S}]+"""
)
'''Our basic word tokenizer regex.'''


def shell_docstring(command, name):
    '''
    Utility for creating docstrings:

      __doc__ += shell_docstring(cli, 'command-name')
    '''
    # Must have a leading newline for Sphinx to process this correctly
    text = '\n'
    text += '**%s**::\n\n  ' % name
    text += click.Context(command, info_name=name)\
                 .get_help()\
                 .replace('\n', '\n  ')
    text += '\n'
    return text


def verbosity(level):
    '''Set logging verbosity to this level (0 to 2 inclusive).
    '''
    LOG_LEVELS = [
        logging.WARNING, logging.INFO, logging.DEBUG
    ]
    logging.basicConfig(
        format='%(levelname)s\t%(message)s',
        level=LOG_LEVELS[min(len(LOG_LEVELS) - 1, level)],
    )


def fdiv_null(a, b):
    '''Return a/b, or ``None`` if b is 0.
    '''
    return None if b == 0 else float(a) / b


def read_jsonlines(filename, nlines=None):
    '''Read an entire stats file into memory, and return it as a list of
    JSON objects.
    '''
    with (sys.stdin if filename == '-' else open(filename, 'r')) as f:
        return list(itertools.islice((json.loads(line.rstrip('\r\n'))
                                      for line in f),
                                     nlines))


def autodetect_log(data, wp, tc, ic):
    '''Detect which game this log contains, and return the appropriate object.
    '''
    log_keys = data[0].keys()
    if 'wordPredictions' in log_keys:
        return wp
    elif 'textCompletions' in log_keys:
        return tc
    elif 'inputCorrections' in log_keys:
        return ic
    else:
        raise ValueError(
            'Unrecognized log line %s' % data[0]
        )


def zip_special(a, b):
    '''A bit like Python's zip, except:
      - Only works for "lengthable" 'a', 'b'.
      - If either is length=0, treat it as a sequence of None.
      - If either 'a' or 'b' is length=1, "broadcast" it to match the other.
      - Fail if the lengths are not =0, =1, or matching.
    '''
    if len(a) == 0:
        return zip(it.repeat(None), b)
    elif len(b) == 0:
        return zip(a, it.repeat(None))
    elif len(a) == 1:
        return zip(it.repeat(a[0]), b)
    elif len(b) == 1:
        return zip(a, it.repeat(b[0]))
    elif len(a) == len(b):
        return zip(a, b)
    else:
        raise ValueError("Length mismatch for zip_special: %d and %d"
                         % (len(a), len(b)))


class JsonParam(click.ParamType):
    '''Click parameter type for parsing JSON.
    If the parameter is a valid filename, assumes that it is a path to a json
    file, and reads that file instead.
    '''
    name = 'json'

    def convert(self, value, param, ctx):
        try:
            if os.path.exists(value):
                with open(value) as f:
                    return json.load(f)
            else:
                return json.loads(value)
        except ValueError as e:
            self.fail(str(e))

    def get_metavar(self, param):
        return 'JSON'


class ParamChoice(click.ParamType):
    '''Like ``click.Choice``, but looks up attributes on specific subclasses.

    To subclass, define:
    ``name` - the descriptive name for the parameter
    ``choices`` - a list of strings, each of which is a valid attr name
    '''

    def convert(self, value, param, ctx):
        if value in type(self).choices:
            return getattr(self, value)
        else:
            self.fail('expected one of {%s}, actually "%s"' % (
                ', '.join(map(repr, type(self).choices)),
                value
            ))

    def get_metavar(self, param):
        return '(%s)' % ('|'.join(type(self).choices))


class TokenFilter(ParamChoice):
    '''Filters for classes of tokens.
    '''
    SPACE_PATTERN = regex.compile(r'\s')
    ALPHA_PATTERN = regex.compile(r'\p{L}')
    NONALPHA_PATTERN = regex.compile(r'[^\p{L}]')
    EMOJI_PATTERN = emoji.get_emoji_regexp()
    MARKER_PATTERN = regex.compile(r'@\p{Lu}*')

    @staticmethod
    def all(t):
        '''Allow all tokens.'''
        return True

    @staticmethod
    def nospace(t):
        '''Exclude tokens that contain whitespace.'''
        return TokenFilter.SPACE_PATTERN.search(t) is None

    @staticmethod
    def nomarker(t):
        '''Exclude tokens that are 'markers' (such as @ALPHANUM).'''
        return TokenFilter.MARKER_PATTERN.match(t) is None

    @staticmethod
    def alpha(t):
        '''Include tokens that contain alphabetic characters.'''
        return TokenFilter.nomarker(t) and \
            (TokenFilter.ALPHA_PATTERN.search(t) is not None)

    @staticmethod
    def alphaonly(t):
        '''Include tokens that are all alphabetic characters.'''
        return TokenFilter.nomarker(t) and \
            (TokenFilter.NONALPHA_PATTERN.search(t) is None)

    @staticmethod
    def alphaemoji(t):
        '''Include tokens that are contain alphabetic characters,
        or an emoji.'''
        return (TokenFilter.nomarker(t) and
                (TokenFilter.ALPHA_PATTERN.search(t) is not None)) or \
            (TokenFilter.EMOJI_PATTERN.search(t) is not None)

    @staticmethod
    def emoji(t):
        '''Include tokens that are an emoji.'''
        return TokenFilter.EMOJI_PATTERN.search(t) is not None

    name = 'token_filter'
    choices = ['all', 'nospace', 'nomarker',
               'alpha', 'alphaonly', 'alphaemoji', 'emoji']


# deprecated & included for legacy API compatibility reasons
def token_filter(name):
    '''Deprecated & included for legacy API compatibility reasons.
    '''
    warnings.warn('deprecated', DeprecationWarning)
    return TokenFilter().convert(name, None, None)


def qualified_name(x):
    '''Return the qualified name of 'x' (including the module).
    The format is: ``module.submodule:attr.subattr``.
    '''
    return '%s:%s' % (x.__module__, x.__qualname__)


QUALIFIED_NAME_PATTERN = re.compile('^([^: ]+):([^: ]+)$')


def is_qualified_name(name):
    '''Determine if this is a qualified name of a type (with module).
    '''
    return QUALIFIED_NAME_PATTERN.match(name) is not None


def lookup_qualified_name(name, base_package=None):
    '''Return the attribute for the qualified 'name'
    (which should be module.submodule:attr.subattr,
    e.g. from `qualified_name`).
    '''
    m = QUALIFIED_NAME_PATTERN.match(name)
    if m is None:
        raise ValueError('could not parse qualified name "%s"' % name)

    module_name = m.group(1)
    attr_names = m.group(2)
    module = importlib.import_module(module_name, base_package)
    try:
        obj = module
        for attr_name in attr_names.split('.'):
            obj = getattr(obj, attr_name)
        return obj
    except AttributeError:
        raise AttributeError(
            'module %s has no attribute %r'
            % (module, attr_names))



from setuptools import setup

with open('version.txt', 'r') as v:
    version = v.readline().strip()

with open('README.md', 'r') as r:
    readme = r.read()

setup(name='lmchallenge',
      version=version,
      description='LM Challenge'
      ' - A library & tools to evaluate predictive language models.',
      long_description=readme,

      url='https://github.com/Microsoft/LMChallenge',
      author='Microsoft Corporation',
      author_email='swiftkey-deep@service.microsoft.com',
      license='MIT',

      packages=['lmchallenge', 'lmchallenge.core'],
      install_requires=[
          'click',
          'emoji',
          'regex',
      ],
      tests_require=[
          'nose',
      ],
      test_suite='nose.collector',
      entry_points='''
      [console_scripts]
      lmc=lmchallenge:cli

      lmc-run=lmchallenge.run:cli
      lmc-stats=lmchallenge.stats:cli
      lmc-ic-opt=lmchallenge.ic_opt:cli
      lmc-pretty=lmchallenge.pretty:cli
      lmc-page=lmchallenge.page:cli
      ''')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

'''Script to optimize the mixing parameters for an ``ic`` log file.
(For the Python API, see ``lmchallenge.stats``.)
'''

import click
import json
from .core import common
from .stats import Ic, Domain


@click.command()
@click.argument('log', type=click.Path(exists=True, dir_okay=False))
@click.option('-v', '--verbose', default=0, count=True,
              help='How much human-readable detail to print to STDERR.')
@click.option('-t', '--target', default='accuracy',
              help='Which metric to optimize for.')
@click.option('-n', '--lines', type=click.INT,
              help='Limit input to this number of lines')
@click.option('-f', '--filter', type=common.TokenFilter(),
              default='alphaemoji',
              help='Only count tokens which match this filter.')
@click.option('-a', '--alpha', type=Domain(),
              default=repr(Ic.DEFAULT_ALPHA),
              help='Grid search domain for alpha (model score mixing weight)')
@click.option('-o', '--oov-penalty', type=Domain(),
              default=repr(Ic.DEFAULT_OOV_PENALTY),
              help='Grid search domain for the oov penalty')
def cli(verbose, log, lines, **args):
    '''Optimize the ic stats, and return the optimal mixing settings,
    as a JSON dictionary.

    Use this with the ``-a`` setting of ``lmc stats`` and ``lmc_pretty`` to
    speed up execution (required for ``lmc_pretty``).
    '''
    common.verbosity(verbose)
    best = Ic.optimize(common.read_jsonlines(log, lines), **args)
    click.echo(json.dumps(best, sort_keys=True))


__doc__ += common.shell_docstring(cli, 'lmc ic-opt')
if __name__ == '__main__':
    cli()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

'''Utility for pretty-printing the model performance from a ``wp``,
``tc``, or ``ic`` log file.
'''

import errno
import click
import io
import collections
from . import stats
from .core import common


# Utilities

class AnsiRender:
    '''A helper for rendering in ANSI color codes'''

    BLACK = 0
    RED = 1
    GREEN = 2
    YELLOW = 3
    BLUE = 4
    MAGENTA = 5
    CYAN = 6
    WHITE = 7
    DEFAULT = 9

    def __init__(self, outf):
        self.f = outf
        self.index = self.DEFAULT
        self.bold = False

    def default(self):
        self.color(self.DEFAULT, False)

    def color(self, index, bold):
        if self.bold and not bold:
            self.f.write(u'\x1b[0;%dm' % (30 + index))
        elif bold and not self.bold:
            self.f.write(u'\x1b[1;%dm' % (30 + index))
        elif self.index != index:
            self.f.write(u'\x1b[%dm' % (30 + index))
        self.index = index
        self.bold = bold

    def write(self, s):
        self.f.write(s)

    def close(self):
        self.f.close()


# Specific challenge programs

class Wp:
    @staticmethod
    def target_rendering(token_info):
        '''Get a string rendering of a token_info (which may be redacted
        as *****).
        '''
        return token_info.get('target') or ('*' * token_info['targetChars'])

    @staticmethod
    def best_rank(ranks, rank_when_missing=None):
        try:
            return min(r for r in ranks if r is not None)
        except ValueError:
            # min of empty sequence
            return rank_when_missing

    def __init__(self, filter):
        self.filter = filter

    def render_pretty(self, row, out):
        '''Render a word-by-word next-word-prediction display
        - top prediction: bold green
        - 2nd/3rd prediction: green
        - any other prediction: yellow
        - not predicted: red
        '''
        for word_info in row['wordPredictions']:
            word = word_info.get('target')
            rank = word_info.get('rank')
            if word is not None and not self.filter(word):
                out.color(AnsiRender.BLUE, False)
            elif rank is None:
                out.color(AnsiRender.RED, False)
            elif rank < 1:
                out.color(AnsiRender.GREEN, True)
            elif rank < 3:
                out.color(AnsiRender.GREEN, False)
            else:
                out.color(AnsiRender.YELLOW, False)
            out.write(Wp.target_rendering(word_info) + ' ')
        out.default()

    def render_diff(self, baselines, line, out):
        '''Render a comparison of a model's rankings against one or more
        baselines.
        '''
        infos = zip(*(x['wordPredictions'] for x in (line,) + baselines))
        for word_info, *baseline_infos in infos:
            target_render = Wp.target_rendering(word_info)
            baseline_target_render = map(Wp.target_rendering, baseline_infos)
            if not all(bt == target_render for bt in baseline_target_render):
                raise ValueError(
                    'sequence of target terms doesn\'t align between'
                    ' evaluation runs being compared')

            rank = word_info.get('rank', 1e10)
            best_baseline_rank = Wp.best_rank(b.get('rank', 1e10)
                                              for b in baseline_infos)
            bold = ((best_baseline_rank <= 1 and rank > 1) or
                    (rank <= 1 and best_baseline_rank > 1) or
                    (best_baseline_rank <= 3 and rank > 3) or
                    (rank <= 3 and best_baseline_rank > 3))
            target = word_info.get('target')
            if target is not None and not self.filter(target):
                out.color(AnsiRender.BLUE, False)
            elif best_baseline_rank < rank:
                out.color(AnsiRender.RED, bold)
            elif rank < best_baseline_rank:
                out.color(AnsiRender.GREEN, bold)
            else:
                out.default()
            out.write(target_render + ' ')
        out.default()


class Tc:
    CharacterState = collections.namedtuple(
        'CharacterState', ['char', 'rank', 'autopredicted']
    )

    @staticmethod
    def get_character_states(row):
        '''Return a flattened sequence of CharacterStates from a result.
        '''
        for token_info in row['textCompletions']:
            target = token_info.get('target')
            rank = token_info.get('rank')
            auto = False
            for ch in (target or [None] * token_info['targetChars']):
                yield Tc.CharacterState(ch, rank, auto)
                auto = True

    def __init__(self, filter):
        self.filter = filter

    def render_pretty(self, row, out):
        '''Render a colored version of a single model's text completion performance.
        - first predicted character: yellow
        - subsequent predicted characters: green
        - not predicted: red
        - ignored: blue
        '''
        for char, rank, autopredicted in Tc.get_character_states(row):
            if char is not None and not self.filter(char):
                out.color(AnsiRender.BLUE, False)
            elif rank is None:
                out.color(AnsiRender.RED, False)
            elif not autopredicted:
                out.color(AnsiRender.YELLOW, False)
            else:
                out.color(AnsiRender.GREEN, False)
            out.write(char or '*')
        out.default()

    def render_diff(self, baselines, line, out):
        partitions = (Tc.get_character_states(x) for x in (line,) + baselines)
        for (char, rank, autopredicted), *baseline_states in zip(*partitions):
            if not all(state.char == char for state in baseline_states):
                raise ValueError(
                    'sequence of characters doesn\'t align between evaluation'
                    ' runs being compared')
            predicted = rank is not None
            predicted_baseline = any(state.rank is not None
                                     for state in baseline_states)
            if char is not None and not self.filter(char):
                out.color(AnsiRender.BLUE, False)
            elif (not predicted) and (not predicted_baseline):
                out.color(AnsiRender.DEFAULT, False)
            elif predicted and predicted_baseline:
                out.color(AnsiRender.CYAN, False)
            elif predicted and (not predicted_baseline):
                out.color(AnsiRender.GREEN, False)
            elif (not predicted) and predicted_baseline:
                out.color(AnsiRender.RED, False)
            else:
                assert False, 'should-be-unreachable code'
            out.write(char or '*')
        out.default()


class Ic:
    def __init__(self, filter):
        self.filter = filter

    def preprocess(self, data, alpha, oov_penalty):
        '''Compute the target ranking with the given mixing settings,
        and store it back in the log.
        Yields a copy of data, with additional word_info:

        ``rank`` - INT - the rank of the correct word
        '''
        for line in data:
            infos = []
            for word_info in line['inputCorrections']:
                out_info = word_info.copy()
                out_info['rank'] = stats.Ic.rank(
                    word_info, alpha=alpha, oov_penalty=oov_penalty
                )
                infos.append(out_info)
            out_line = line.copy()
            out_line['inputCorrections'] = infos
            yield out_line

    def render_pretty(self, line, out):
        '''Render a word-by-word correction display:

        +-----------+-----------+--------+
        | Before    | After     | Color  |
        +===========+===========+========+
        | Incorrect | Incorrect | Yellow |
        +-----------+-----------+--------+
        | Incorrect | Correct   | Green  |
        +-----------+-----------+--------+
        | Correct   | Incorrect | Red    |
        +-----------+-----------+--------+
        | Correct   | Correct   | White  |
        +-----------+-----------+--------+
        |        (Skip)         | Blue   |
        +-----------+-----------+--------+
        '''
        for word_info in line['inputCorrections']:
            word = word_info.get('target')
            before_after = (
                stats.Ic.before_pass(word_info),
                word_info['rank'] == 1
            )
            if word is not None and not self.filter(word):
                out.color(AnsiRender.BLUE, False)
            elif before_after == (False, False):
                out.color(AnsiRender.YELLOW, False)
            elif before_after == (False, True):
                out.color(AnsiRender.GREEN, False)
            elif before_after == (True, False):
                out.color(AnsiRender.RED, False)
            else:  # (True, True)
                out.color(AnsiRender.DEFAULT, False)
            out.write(Wp.target_rendering(word_info) + ' ')
        out.default()

    def render_diff(self, baselines, line, out):
        '''Render the difference between the best of 'baselines'
        (a tuple of lines) and 'line'.

        Examples are rendered as follows. Where the color is marked with a
        star, and the original was correct, the output is marked bold.

        +-----------+-----------+--------++
        | Baseline  | Actual    | Color   |
        +===========+===========+=========+
        | Incorrect | Incorrect | Yellow* |
        +-----------+-----------+---------+
        | Incorrect | Correct   | Green*  |
        +-----------+-----------+---------+
        | Correct   | Incorrect | Red*    |
        +-----------+-----------+---------+
        | Correct   | Correct   | White   |
        +-----------+-----------+---------+
        |        (Skip)         | Blue    |
        +-----------+-----------+---------+
        '''
        infos = zip(*(x['inputCorrections'] for x in (line,) + baselines))
        for word_info, *baseline_word_infos in infos:
            before = stats.Ic.before_pass(word_info)
            baseline_actual = (
                1 == min(x['rank'] for x in baseline_word_infos),
                1 == word_info['rank']
            )
            word = word_info.get('target')
            if word is not None and not self.filter(word):
                out.color(AnsiRender.BLUE, False)
            elif baseline_actual == (False, False):
                out.color(AnsiRender.YELLOW, before)
            elif baseline_actual == (False, True):
                out.color(AnsiRender.GREEN, before)
            elif baseline_actual == (True, False):
                out.color(AnsiRender.RED, before)
            else:  # (True, True)
                out.color(AnsiRender.DEFAULT, False)
            out.write(Wp.target_rendering(word_info) + ' ')
        out.default()


# Generic pretty printing

def pretty(*data, filter=common.TokenFilter.all, opt_args=None):
    '''Compute and return the pretty-printed difference between ``data`` and
    ``baseline_data`` (if specified, otherwise just the absolute performance of
    ``data``.)

    ``data`` - a list of (1+) lists of log lines
               (e.g. each from ``common.read_jsonlines``)

    ``returns`` - a generator of strings containing ANSI color codes
    '''
    pretty_program = common.autodetect_log(data[0], wp=Wp, tc=Tc, ic=Ic)(
        filter=filter
    )
    if opt_args:
        data = [pretty_program.preprocess(d, **args)
                for d, args in common.zip_special(data, opt_args)]

    for lines in zip(*data):
        out = AnsiRender(io.StringIO())
        if 2 <= len(lines):
            pretty_program.render_diff(tuple(lines[:-1]), lines[-1], out)
        else:
            pretty_program.render_pretty(lines[0], out)
        yield out.f.getvalue()


@click.command()
@click.argument('log', nargs=-1, type=click.Path(exists=True, dir_okay=False))
@click.option('-f', '--filter', type=common.TokenFilter(),
              default='alphaemoji',
              help='Only count tokens which match this filter.')
@click.option('-a', '--opt-args', type=common.JsonParam(), multiple=True,
              help='Pass these arguments to the pretty program (for ``ic``),'
              ' for example pre-computed using ``lmc ic-opt``.'
              ' If comparing to baselines, we expect multiple arguments - the'
              ' same number as the number of log files passed.')
def cli(log, **args):
    '''Pretty-print results from an LMChallenge game (using ANSI color codes).

    If multiple files are specified, use all but the last as 'baselines',
    then pretty-print the differences against the best of a set of baseline
    results (the best at each individual location is used).

    Otherwise pretty-print the basic quality of results.
    '''
    if len(log) == 0:
        log = ['-']
    try:
        for line in pretty(*map(common.read_jsonlines, log), **args):
            # We don't want click.echo's clever handling of colors - we want to
            # always include them
            print(line)
    except IOError as e:
        if e.errno == errno.EPIPE:
            pass


__doc__ += common.shell_docstring(cli, 'lmc pretty')
if __name__ == '__main__':
    cli()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

import functools


# Helpers

def _merge_dicts(*dicts):
    m = {}
    for d in dicts:
        m.update(d)
    return m


# Basic definition

class Monoid:
    '''A wrapper for a simple triple of operations - identity, lift, and combine.
    These should obey the Monoid laws of associativity and identity, i.e.

        combine(lift(x), identity) == lift(x)
        combine(identity, lift(x)) == lift(x)
        combine(lift(x), combine(lift(y), lift(z)) \
            == combine(combine(lift(x), lift(y)), lift(z))
    '''

    __slots__ = ['identity', 'lift', 'combine']

    def __init__(self, identity, lift, combine):
        self.identity = identity
        self.lift = lift
        self.combine = combine

    def reduce(self, xs):
        '''Lift each x using the Monoid, then combine them all together.
        '''
        return functools.reduce(self.combine,
                                (self.lift(x) for x in xs),
                                self.identity)


# Simple monoids

counting = Monoid(identity=0, lift=lambda x: 1, combine=lambda a, b: a + b)

summing = Monoid(identity=0, lift=lambda x: x, combine=lambda a, b: a + b)

maxing = Monoid(
    identity=None,
    lift=lambda x: x,
    combine=lambda a, b: b if a is None else (None if b is None else max(a, b))
)

uniqing = Monoid(
    identity=set([]),
    lift=lambda x: set([x]),
    combine=lambda a, b: a.union(b)
)


# Monoid transformers

def map_input(f, monoid):
    '''Map the input to a monoid by a preparing function.
    '''
    return Monoid(
        monoid.identity,
        lambda x: monoid.lift(f(x)),
        monoid.combine
    )


def filter_input(pred, monoid):
    '''Filter inputs to a monoid using a predicate (applied to the un-lifted
    input).
    '''
    return Monoid(
        monoid.identity,
        lambda x: monoid.lift(x) if pred(x) else monoid.identity,
        monoid.combine
    )


def merge(*submonoids):
    '''For monoids of 'output type' `dict`, create a `Monoid` that merges
    their keys (which should not collide).
    '''
    return Monoid(
        _merge_dicts(*[m.identity for m in submonoids]),
        lambda x: _merge_dicts(*[m.lift(x) for m in submonoids]),
        lambda a, b: _merge_dicts(*[m.combine(a, b) for m in submonoids])
    )


def keys(**submonoids):
    '''Return a monoid that produces {k: submonoid} mappings.
    '''
    return Monoid(
        {k: m.identity for k, m in submonoids.items()},
        lambda x: {k: m.lift(x) for k, m in submonoids.items()},
        lambda a, b: {k: m.combine(a[k], b[k])
                      for k, m in submonoids.items()}
    )


def multiplex(dispatch, **submonoids):
    '''Return a monoid that selects which submonoid to send an to based on a
    dispatch function.
    '''
    def lift(x):
        k = dispatch(x)
        {k: submonoids[k].lift(x)}
    return Monoid(
        {k: m.identity for k, m in submonoids.items()},
        lift,
        lambda a, b: {k: m.combine(a.get(k, m.identity), b.get(k, m.identity))
                      for k, m in submonoids.items()}
    )
